@@ -2,11 +2,13 @@ import concurrent.futures
 import logging
 import os
 import threading
+from typing import Optional
 
 import requests
 from cleo.application import Application as BaseApplication
 from cleo.commands.command import Command
 from cleo.formatters.style import Style
+from cleo.helpers import argument, option
 from click import launch
 
 from streamrip import __version__
@@ -18,6 +20,7 @@ logging.basicConfig(level="WARNING")
 logger = logging.getLogger("streamrip")
 
 outdated = False
+newest_version = __version__
 
 
 class DownloadCommand(Command):
@@ -34,7 +37,7 @@ class DownloadCommand(Command):
 
     help = (
         "\nDownload <title>Dreams</title> by <title>Fleetwood Mac</title>:\n"
-        "$ <cmd>rip url https://www.deezer.com/en/track/63480987</cmd>\n\n"
+        "$ <cmd>rip url https://www.deezer.com/us/track/67549262</cmd>\n\n"
         "Batch download urls from a text file named <path>urls.txt</path>:\n"
         "$ <cmd>rip url --file urls.txt</cmd>\n\n"
         "For more information on Quality IDs, see\n"
@@ -43,9 +46,11 @@ class DownloadCommand(Command):
 
     def handle(self):
         global outdated
+        global newest_version
 
         # Use a thread so that it doesn't slow down startup
         update_check = threading.Thread(target=is_outdated, daemon=True)
+        update_check.start()
 
         config = Config()
         path, codec, quality, no_db = clean_options(
@@ -83,16 +88,39 @@ class DownloadCommand(Command):
         elif not urls and path is None:
             self.line("<error>Must pass arguments. See </><cmd>rip url -h</cmd>.")
 
-        try:
-            update_check.join()
-            if outdated:
-                self.line(
-                    "<info>A new version of streamrip is available! Run</info> "
-                    "<cmd>pip3 install streamrip --upgrade to update</cmd>"
-                )
-        except RuntimeError as e:
-            logger.debug("Update check error: %s", e)
-            pass
+        update_check.join()
+        if outdated:
+            import re
+            import subprocess
+
+            self.line(
+                f"<info>Updating streamrip to <title>v{newest_version}</title>...</info>\n"
+            )
+
+            # update in background
+            update_p = subprocess.Popen(
+                ["pip3", "install", "streamrip", "--upgrade"],
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+            )
+
+            md_header = re.compile(r"#\s+(.+)")
+            bullet_point = re.compile(r"-\s+(.+)")
+            code = re.compile(r"`([^`]+)`")
+            issue_reference = re.compile(r"(#\d+)")
+
+            release_notes = requests.get(
+                "https://api.github.com/repos/nathom/streamrip/releases/latest"
+            ).json()["body"]
+
+            release_notes = md_header.sub(r"<header>\1</header>", release_notes)
+            release_notes = bullet_point.sub(r"<options=bold>•</> \1", release_notes)
+            release_notes = code.sub(r"<cmd>\1</cmd>", release_notes)
+            release_notes = issue_reference.sub(r"<options=bold>\1</>", release_notes)
+
+            self.line(release_notes)
+
+            update_p.wait()
 
         return 0
 
@@ -131,11 +159,12 @@ class SearchCommand(Command):
 
 class DiscoverCommand(Command):
     """
-    Browse and download items in interactive mode (Qobuz only).
+    Browse and download items in interactive mode (Qobuz and Deezer only).
 
     discover
-        {--s|scrape : Download all of the items in the list}
+        {--scrape : Download all of the items in the list}
         {--m|max-items=50 : The number of items to fetch}
+        {--s|source=qobuz : The source to download from (<cmd>qobuz</cmd> or <cmd>deezer</cmd>)}
         {list=ideal-discography : The list to fetch}
     """
 
@@ -144,7 +173,7 @@ class DiscoverCommand(Command):
         "$ <cmd>rip discover</cmd>\n\n"
         "Browse the best-sellers list\n"
         "$ <cmd>rip discover best-sellers</cmd>\n\n"
-        "Available options for <info>list</info>:\n\n"
+        "Available options for Qobuz <cmd>list</cmd>:\n\n"
         "    • most-streamed\n"
         "    • recent-releases\n"
         "    • best-sellers\n"
@@ -159,19 +188,42 @@ class DiscoverCommand(Command):
         "    • universal-classic\n"
         "    • universal-jazz\n"
         "    • universal-jeunesse\n"
-        "    • universal-chanson\n"
+        "    • universal-chanson\n\n"
+        "Browse the Deezer editorial releases list\n"
+        "$ <cmd>rip discover --source deezer</cmd>\n\n"
+        "Browse the Deezer charts\n"
+        "$ <cmd>rip discover --source deezer charts</cmd>\n\n"
+        "Available options for Deezer <cmd>list</cmd>:\n\n"
+        "    • releases\n"
+        "    • charts\n"
+        "    • selection\n"
     )
 
     def handle(self):
-        from streamrip.constants import QOBUZ_FEATURED_KEYS
-
-        chosen_list = self.argument("list")
+        source = self.option("source")
         scrape = self.option("scrape")
+        chosen_list = self.argument("list")
         max_items = self.option("max-items")
 
-        if chosen_list not in QOBUZ_FEATURED_KEYS:
-            self.line(f'<error>Error: list "{chosen_list}" not available</error>')
-            self.line(self.help)
+        if source == "qobuz":
+            from streamrip.constants import QOBUZ_FEATURED_KEYS
+
+            if chosen_list not in QOBUZ_FEATURED_KEYS:
+                self.line(f'<error>Error: list "{chosen_list}" not available</error>')
+                self.line(self.help)
+                return 1
+        elif source == "deezer":
+            from streamrip.constants import DEEZER_FEATURED_KEYS
+
+            if chosen_list not in DEEZER_FEATURED_KEYS:
+                self.line(f'<error>Error: list "{chosen_list}" not available</error>')
+                self.line(self.help)
+                return 1
+
+        else:
+            self.line(
+                "<error>Invalid source. Choose either <cmd>qobuz</cmd> or <cmd>deezer</cmd></error>"
+            )
             return 1
 
         config = Config()
@@ -183,12 +235,14 @@ class DiscoverCommand(Command):
             return 0
 
         if core.interactive_search(
-            chosen_list, "qobuz", "featured", limit=int(max_items)
+            chosen_list, source, "featured", limit=int(max_items)
         ):
             core.download()
         else:
             self.line("<error>No items chosen, exiting.</error>")
 
+        return 0
+
 
 class LastfmCommand(Command):
     """
@@ -232,16 +286,19 @@ class ConfigCommand(Command):
         {--qobuz : Set the credentials for Qobuz}
         {--tidal : Log into Tidal}
         {--deezer : Set the Deezer ARL}
+        {--music-app : Configure the config file for usage with the macOS Music App}
         {--reset : Reset the config file}
         {--update : Reset the config file, keeping the credentials}
     """
 
+    _config: Optional[Config]
+
     def handle(self):
         import shutil
 
         from .constants import CONFIG_DIR, CONFIG_PATH
 
-        config = Config()
+        self._config = Config()
 
         if self.option("path"):
             self.line(f"<info>{CONFIG_PATH}</info>")
@@ -251,10 +308,10 @@ class ConfigCommand(Command):
             launch(CONFIG_PATH)
 
         if self.option("reset"):
-            config.reset()
+            self._config.reset()
 
         if self.option("update"):
-            config.update()
+            self._config.update()
 
         if self.option("open-vim"):
             if shutil.which("nvim") is not None:
@@ -271,18 +328,85 @@ class ConfigCommand(Command):
 
             client = TidalClient()
             client.login()
-            config.file["tidal"].update(client.get_tokens())
-            config.save()
+            self._config.file["tidal"].update(client.get_tokens())
+            self._config.save()
             self.line("<info>Credentials saved to config.</info>")
 
         if self.option("deezer"):
+            from streamrip.clients import DeezerClient
+            from streamrip.exceptions import AuthenticationError
+
             self.line(
                 "Follow the instructions at <url>https://github.com"
                 "/nathom/streamrip/wiki/Finding-your-Deezer-ARL-Cookie</url>"
             )
 
-            config.file["deezer"]["arl"] = self.ask("Paste your ARL here: ")
-            config.save()
+            given_arl = self.ask("Paste your ARL here: ").strip()
+            self.line("<comment>Validating arl...</comment>")
+
+            try:
+                DeezerClient().login(arl=given_arl)
+                self._config.file["deezer"]["arl"] = given_arl
+                self._config.save()
+                self.line("<b>Sucessfully logged in!</b>")
+
+            except AuthenticationError:
+                self.line("<error>Could not log in. Double check your ARL</error>")
+
+        if self.option("qobuz"):
+            import getpass
+            import hashlib
+
+            self._config.file["qobuz"]["email"] = self.ask("Qobuz email:")
+            self._config.file["qobuz"]["password"] = hashlib.md5(
+                getpass.getpass("Qobuz password (won't show on screen): ").encode()
+            ).hexdigest()
+            self._config.save()
+
+        if self.option("music-app"):
+            self._conf_music_app()
+
+    def _conf_music_app(self):
+        import subprocess
+        import xml.etree.ElementTree as ET
+        from pathlib import Path
+        from tempfile import mktemp
+
+        # Find the Music library folder
+        temp_file = mktemp()
+        music_pref_plist = Path(Path.home()) / Path(
+            "Library/Preferences/com.apple.Music.plist"
+        )
+        # copy preferences to tempdir
+        subprocess.run(["cp", music_pref_plist, temp_file])
+        # convert binary to xml for parsing
+        subprocess.run(["plutil", "-convert", "xml1", temp_file])
+        items = iter(ET.parse(temp_file).getroot()[0])
+
+        for item in items:
+            if item.text == "NSNavLastRootDirectory":
+                break
+
+        library_folder = Path(next(items).text)
+        os.remove(temp_file)
+
+        # cp ~/library/preferences/com.apple.music.plist music.plist
+        # plutil -convert xml1 music.plist
+        # cat music.plist | pbcopy
+
+        self._config.file["downloads"]["folder"] = os.path.join(
+            library_folder, "Automatically Add to Music.localized"
+        )
+
+        conversion_config = self._config.file["conversion"]
+        conversion_config["enabled"] = True
+        conversion_config["codec"] = "ALAC"
+        conversion_config["sampling_rate"] = 48000
+        conversion_config["bit_depth"] = 24
+
+        self._config.file["filepaths"]["folder_format"] = ""
+        self._config.file["artwork"]["keep_hires_cover"] = False
+        self._config.save()
 
 
 class ConvertCommand(Command):
@@ -360,7 +484,8 @@ class ConvertCommand(Command):
                     futures.append(
                         executor.submit(
                             ConverterCls(
-                                filename=os.path.join(dirname, file), **converter_args
+                                filename=os.path.join(dirname, file),
+                                **converter_args,
                             ).convert
                         )
                     )
@@ -379,7 +504,8 @@ class ConvertCommand(Command):
             ConverterCls(filename=path, **converter_args).convert()
         else:
             self.line(
-                f'<error>Path <path>"{path}"</path> does not exist.</error>', fg="red"
+                f'<error>Path <path>"{path}"</path> does not exist.</error>',
+                fg="red",
             )
 
 
@@ -394,11 +520,67 @@ class RepairCommand(Command):
     help = "\nRetry up to 20 failed downloads\n$ <cmd>rip repair --max-items 20</cmd>\n"
 
     def handle(self):
-        max_items = clean_options(self.option("repair"))
+        max_items = next(clean_options(self.option("max-items")))
         config = Config()
         RipCore(config).repair(max_items=max_items)
 
 
+class DatabaseCommand(Command):
+    """
+    View and manage rip's databases.
+
+    db
+        {name : <cmd>downloads</cmd> or <cmd>failed-downloads</cmd>.}
+        {--l|list : Display the contents of the database.}
+        {--reset : Reset the database.}
+    """
+
+    _table_style = "box-double"
+
+    def handle(self) -> None:
+        from . import db
+        from .config import Config
+
+        config = Config()
+        db_name = self.argument("name").replace("-", "_")
+
+        self._path = config.file["database"][db_name]["path"]
+        self._db = db.CLASS_MAP[db_name](self._path)
+
+        if self.option("list"):
+            getattr(self, f"_render_{db_name}")()
+
+        if self.option("reset"):
+            os.remove(self._path)
+
+    def _render_downloads(self):
+        from cleo.ui.table import Table
+
+        id_table = Table(self._io)
+        id_table.set_style(self._table_style)
+        id_table.set_header_title("IDs")
+        id_table.set_headers(list(self._db.structure.keys()))
+        id_table.add_rows(id for id in iter(self._db) if id[0].isalnum())
+        id_table.render()
+
+        url_table = Table(self._io)
+        url_table.set_style(self._table_style)
+        url_table.set_header_title("URLs")
+        url_table.set_headers(list(self._db.structure.keys()))
+        url_table.add_rows(id for id in iter(self._db) if not id[0].isalnum())
+        url_table.render()
+
+    def _render_failed_downloads(self):
+        from cleo.ui.table import Table
+
+        id_table = Table(self._io)
+        id_table.set_style(self._table_style)
+        id_table.set_header_title("Failed Downloads")
+        id_table.set_headers(["Source", "Media Type", "ID"])
+        id_table.add_rows(iter(self._db))
+        id_table.render()
+
+
 STRING_TO_PRIMITIVE = {
     "None": None,
     "True": True,
@@ -412,7 +594,12 @@ class Application(BaseApplication):
 
     def _run(self, io):
         if io.is_debug():
+            from .constants import CONFIG_DIR
+
             logger.setLevel(logging.DEBUG)
+            fh = logging.FileHandler(os.path.join(CONFIG_DIR, "streamrip.log"))
+            fh.setLevel(logging.DEBUG)
+            logger.addHandler(fh)
 
         super()._run(io)
 
@@ -424,6 +611,7 @@ class Application(BaseApplication):
         formatter.set_style("path", Style("green", options=["bold"]))
         formatter.set_style("cmd", Style("magenta"))
         formatter.set_style("title", Style("yellow", options=["bold"]))
+        formatter.set_style("header", Style("yellow", options=["bold", "underline"]))
         io.output.set_formatter(formatter)
         io.error_output.set_formatter(formatter)
 
@@ -467,8 +655,10 @@ def clean_options(*opts):
 
 def is_outdated():
     global outdated
+    global newest_version
     r = requests.get("https://pypi.org/pypi/streamrip/json").json()
-    outdated = r["info"]["version"] != __version__
+    newest_version = r["info"]["version"]
+    outdated = newest_version != __version__
 
 
 def main():
@@ -480,6 +670,7 @@ def main():
     application.add(ConfigCommand())
     application.add(ConvertCommand())
     application.add(RepairCommand())
+    application.add(DatabaseCommand())
     application.run()
 
 
