@@ -5,6 +5,7 @@ import html
 import logging
 import os
 import re
+import threading
 from getpass import getpass
 from hashlib import md5
 from string import Formatter
@@ -25,6 +26,7 @@ from streamrip.clients import (
 from streamrip.constants import MEDIA_TYPES
 from streamrip.exceptions import (
     AuthenticationError,
+    IneligibleError,
     ItemExists,
     MissingCredentials,
     NonStreamable,
@@ -211,23 +213,24 @@ class RipCore(list):
         session = self.config.session
         logger.debug(session)
         # So that the dictionary isn't searched for the same keys multiple times
-        artwork, conversion, filepaths = tuple(
-            session[key] for key in ("artwork", "conversion", "filepaths")
+        artwork, conversion, filepaths, metadata = (
+            session[key] for key in ("artwork", "conversion", "filepaths", "metadata")
         )
         concurrency = session["downloads"]["concurrency"]
         return {
+            "restrict_filenames": filepaths["restrict_characters"],
             "parent_folder": session["downloads"]["folder"],
             "folder_format": filepaths["folder_format"],
             "track_format": filepaths["track_format"],
-            "embed_cover": session["artwork"]["embed"],
+            "embed_cover": artwork["embed"],
             "embed_cover_size": artwork["size"],
             "keep_hires_cover": artwork["keep_hires_cover"],
-            "set_playlist_to_album": session["metadata"]["set_playlist_to_album"],
+            "set_playlist_to_album": metadata["set_playlist_to_album"],
             "stay_temp": conversion["enabled"],
             "conversion": conversion,
             "concurrent_downloads": concurrency["enabled"],
             "max_connections": concurrency["max_connections"],
-            "new_tracknumbers": session["metadata"]["new_playlist_tracknumbers"],
+            "new_tracknumbers": metadata["new_playlist_tracknumbers"],
             "download_videos": session["tidal"]["download_videos"],
             "download_booklets": session["qobuz"]["download_booklets"],
             "download_youtube_videos": session["youtube"]["download_videos"],
@@ -237,6 +240,7 @@ class RipCore(list):
             "add_singles_to_folder": filepaths["add_singles_to_folder"],
             "max_artwork_width": int(artwork["max_width"]),
             "max_artwork_height": int(artwork["max_height"]),
+            "exclude_tags": metadata["exclude"],
         }
 
     def repair(self, max_items=None):
@@ -247,6 +251,7 @@ class RipCore(list):
         if max_items is None:
             max_items = float("inf")
 
+        self.db = db.Downloads(None, dummy=True)
         if self.failed_db.is_dummy:
             secho(
                 "Failed downloads database must be enabled in the config file "
@@ -330,7 +335,7 @@ class RipCore(list):
                     self.db.add(str(item_id))
 
             if isinstance(item, Track):
-                item.tag()
+                item.tag(exclude_tags=arguments["exclude_tags"])
                 if arguments["conversion"]["enabled"]:
                     item.convert(**arguments["conversion"])
 
@@ -367,12 +372,12 @@ class RipCore(list):
         if client.source == "deezer" and creds["arl"] == "":
             if self.config.session["deezer"]["deezloader_warnings"]:
                 secho(
-                    "Falling back to Deezloader (max 320kbps MP3). If you have a subscription, run ",
+                    "Falling back to Deezloader (unstable). If you have a subscription, run ",
                     nl=False,
                     fg="yellow",
                 )
                 secho("rip config --deezer ", nl=False, bold=True)
-                secho("to download FLAC files.\n\n", fg="yellow")
+                secho("to log in.", fg="yellow")
             raise DeezloaderFallback
 
         while True:
@@ -385,9 +390,16 @@ class RipCore(list):
                 creds = self.config.creds(client.source)
             except MissingCredentials:
                 logger.debug("Credentials are missing. Prompting..")
+                get_tokens = threading.Thread(
+                    target=client._get_app_id_and_secrets, daemon=True
+                )
+                get_tokens.start()
+
                 self.prompt_creds(client.source)
                 creds = self.config.creds(client.source)
 
+                get_tokens.join()
+
         if (
             client.source == "qobuz"
             and not creds.get("secrets")
@@ -398,6 +410,17 @@ class RipCore(list):
                 self.config.file["qobuz"]["secrets"],
             ) = client.get_tokens()
             self.config.save()
+        elif (
+            client.source == "soundcloud"
+            and not creds.get("client_id")
+            and not creds.get("app_version")
+        ):
+            (
+                self.config.file["soundcloud"]["client_id"],
+                self.config.file["soundcloud"]["app_version"],
+            ) = client.get_tokens()
+            self.config.save()
+
         elif client.source == "tidal":
             self.config.file["tidal"].update(client.get_tokens())
             self.config.save()  # only for the expiry stamp
@@ -406,14 +429,14 @@ class RipCore(list):
         """Return the type of the url and the id.
 
         Compatible with urls of the form:
-            https://www.qobuz.com/us-en/{type}/{name}/{id}
-            https://open.qobuz.com/{type}/{id}
-            https://play.qobuz.com/{type}/{id}
+            https://www.qobuz.com/us-en/type/name/id
+            https://open.qobuz.com/type/id
+            https://play.qobuz.com/type/id
 
-            https://www.deezer.com/us/{type}/{id}
-            https://tidal.com/browse/{type}/{id}
+            https://www.deezer.com/us/type/id
+            https://tidal.com/browse/type/id
 
-        :raises exceptions.ParsingError
+        :raises exceptions.ParsingError:
         """
         parsed: List[Tuple[str, str, str]] = []
 
@@ -442,16 +465,24 @@ class RipCore(list):
                 ("deezer", *extract_deezer_dynamic_link(url)) for url in dynamic_urls
             )
 
-        parsed.extend(URL_REGEX.findall(url))  # Qobuz, Tidal, Dezer
+        parsed.extend(URL_REGEX.findall(url))  # Qobuz, Tidal, Deezer
         soundcloud_urls = SOUNDCLOUD_URL_REGEX.findall(url)
-        soundcloud_items = [self.clients["soundcloud"].get(u) for u in soundcloud_urls]
 
-        parsed.extend(
-            ("soundcloud", item["kind"], url)
-            for item, url in zip(soundcloud_items, soundcloud_urls)
-        )
+        if soundcloud_urls:
+            soundcloud_client = self.get_client("soundcloud")
+            assert isinstance(soundcloud_client, SoundCloudClient)  # for typing
 
-        logger.debug(f"Parsed urls: {parsed}")
+            # TODO: Make this async
+            soundcloud_items = (
+                soundcloud_client.resolve_url(u) for u in soundcloud_urls
+            )
+
+            parsed.extend(
+                ("soundcloud", item["kind"], str(item["id"]))
+                for item in soundcloud_items
+            )
+
+        logger.debug("Parsed urls: %s", parsed)
 
         return parsed
 
@@ -488,6 +519,14 @@ class RipCore(list):
             self._config_corrupted_message(err)
             exit()
 
+        # Do not include tracks that have (re)mix, live, karaoke in their titles
+        # within parentheses or brackets
+        # This will match somthing like "Test (Person Remix]" though, so its not perfect
+        banned_words_plain = re.compile(r"(?i)(?:(?:re)?mix|live|karaoke)")
+        banned_words = re.compile(
+            r"(?i)[\(\[][^\)\]]*?(?:(?:re)?mix|live|karaoke)[^\)\]]*[\]\)]"
+        )
+
         def search_query(title, artist, playlist) -> bool:
             """Search for a query and add the first result to playlist.
 
@@ -503,7 +542,16 @@ class RipCore(list):
                     query = QUERY_FORMAT[lastfm_source].format(
                         title=title, artist=artist
                     )
-                    track = next(self.search(source, query, media_type="track"))
+                    query_is_clean = banned_words_plain.search(query) is None
+
+                    search_results = self.search(source, query, media_type="track")
+                    track = next(search_results)
+
+                    if query_is_clean:
+                        while banned_words.search(track["title"]) is not None:
+                            logger.debug("Track title banned for query=%s", query)
+                            track = next(search_results)
+
                     # Because the track is searched as a single we need to set
                     # this manually
                     track.part_of_tracklist = True
@@ -594,8 +642,17 @@ class RipCore(list):
         logger.debug("searching for %s", query)
 
         client = self.get_client(source)
+
+        if isinstance(client, DeezloaderClient) and media_type == "featured":
+            raise IneligibleError(
+                "Must have premium Deezer account to access editorial lists."
+            )
+
         results = client.search(query, media_type)
 
+        if media_type == "featured":
+            media_type = "album"
+
         if isinstance(results, Generator):  # QobuzClient
             for page in results:
                 tracklist = (
@@ -611,16 +668,20 @@ class RipCore(list):
                     if i >= limit - 1:
                         return
         else:
-            logger.debug("Not generator")
             items = (
-                results.get("data") or results.get("items") or results.get("collection")
+                results.get("data")
+                or results.get("items")
+                or results.get("collection")
+                or results.get("albums", {}).get("data", False)
             )
-            if items is None:
+
+            if not items:
                 raise NoResultsFound(query)
+
             logger.debug("Number of results: %d", len(items))
 
             for i, item in enumerate(items):
-                logger.debug(item["title"])
+                logger.debug(item)
                 yield MEDIA_CLASS[media_type].from_api(item, client)  # type: ignore
                 if i >= limit - 1:
                     return
