@@ -2,13 +2,13 @@
 
 import base64
 import binascii
+import concurrent.futures
 import hashlib
 import json
 import logging
 import re
 import time
 from abc import ABC, abstractmethod
-from pprint import pformat
 from typing import Any, Dict, Generator, Optional, Sequence, Tuple, Union
 
 import deezer
@@ -21,12 +21,9 @@ from .constants import (
     DEEZER_BASE,
     DEEZER_DL,
     DEEZER_FORMATS,
-    DEEZER_MAX_Q,
     QOBUZ_BASE,
     QOBUZ_FEATURED_KEYS,
-    SOUNDCLOUD_APP_VERSION,
     SOUNDCLOUD_BASE,
-    SOUNDCLOUD_CLIENT_ID,
     SOUNDCLOUD_USER_ID,
     TIDAL_AUTH_URL,
     TIDAL_BASE,
@@ -118,6 +115,7 @@ class QobuzClient(Client):
         :type pwd: str
         :param kwargs: app_id: str, secrets: list, return_secrets: bool
         """
+        # TODO: make this faster
         secho(f"Logging into {self.source}", fg="green")
         email: str = kwargs["email"]
         pwd: str = kwargs["pwd"]
@@ -128,23 +126,20 @@ class QobuzClient(Client):
             logger.debug("Already logged in")
             return
 
-        if (kwargs.get("app_id") or kwargs.get("secrets")) in (None, [], ""):
-            secho("Fetching tokens — this may take a few seconds.", fg="magenta")
-            logger.info("Fetching tokens from Qobuz")
-            spoofer = Spoofer()
-            kwargs["app_id"] = spoofer.get_app_id()
-            kwargs["secrets"] = spoofer.get_secrets()
-
-        self.app_id = str(kwargs["app_id"])  # Ensure it is a string
-        self.secrets = kwargs["secrets"]
-
-        self.session = gen_threadsafe_session(
-            headers={"User-Agent": AGENT, "X-App-Id": self.app_id}
-        )
+        if not kwargs.get("app_id") or not kwargs.get("secrets"):
+            self._get_app_id_and_secrets()  # can be async
+        else:
+            self.app_id, self.secrets = (
+                str(kwargs["app_id"]),
+                kwargs["secrets"],
+            )
+            self.session = gen_threadsafe_session(
+                headers={"User-Agent": AGENT, "X-App-Id": self.app_id}
+            )
+            self._validate_secrets()
 
         self._api_login(email, pwd)
         logger.debug("Logged into Qobuz")
-        self._validate_secrets()
         logger.debug("Qobuz client is ready to use")
 
         self.logged_in = True
@@ -217,6 +212,21 @@ class QobuzClient(Client):
 
     # ---------- Private Methods ---------------
 
+    def _get_app_id_and_secrets(self):
+        if not hasattr(self, "app_id") or not hasattr(self, "secrets"):
+            spoofer = Spoofer()
+            self.app_id, self.secrets = (
+                str(spoofer.get_app_id()),
+                spoofer.get_secrets(),
+            )
+
+        if not hasattr(self, "sec"):
+            if not hasattr(self, "session"):
+                self.session = gen_threadsafe_session(
+                    headers={"User-Agent": AGENT, "X-App-Id": self.app_id}
+                )
+            self._validate_secrets()
+
     def _gen_pages(self, epoint: str, params: dict) -> Generator:
         """When there are multiple pages of results, this yields them.
 
@@ -248,11 +258,17 @@ class QobuzClient(Client):
 
     def _validate_secrets(self):
         """Check if the secrets are usable."""
-        for secret in self.secrets:
-            if self._test_secret(secret):
-                self.sec = secret
-                logger.debug("Working secret and app_id: %s - %s", secret, self.app_id)
-                break
+        with concurrent.futures.ThreadPoolExecutor() as executor:
+            futures = [
+                executor.submit(self._test_secret, secret) for secret in self.secrets
+            ]
+
+            for future in concurrent.futures.as_completed(futures):
+                result = future.result()
+                if result is not None:
+                    self.sec = result
+                    break
+
         if not hasattr(self, "sec"):
             raise InvalidAppSecretError(f"Invalid secrets: {self.secrets}")
 
@@ -385,7 +401,7 @@ class QobuzClient(Client):
         else:
             raise InvalidAppSecretError("Cannot find app secret")
 
-        quality = int(get_quality(quality, self.source))
+        quality = int(get_quality(quality, self.source))  # type: ignore
         r_sig = f"trackgetFileUrlformat_id{quality}intentstreamtrack_id{track_id}{unix_ts}{secret}"
         logger.debug("Raw request signature: %s", r_sig)
         r_sig_hashed = hashlib.md5(r_sig.encode("utf-8")).hexdigest()
@@ -422,7 +438,7 @@ class QobuzClient(Client):
             logger.error("Problem getting JSON. Status code: %s", r.status_code)
             raise
 
-    def _test_secret(self, secret: str) -> bool:
+    def _test_secret(self, secret: str) -> Optional[str]:
         """Test the authenticity of a secret.
 
         :param secret:
@@ -431,10 +447,10 @@ class QobuzClient(Client):
         """
         try:
             self._api_get_file_url("19512574", sec=secret)
-            return True
+            return secret
         except InvalidAppSecretError as error:
             logger.debug("Test for %s secret didn't work: %s", secret, error)
-            return False
+            return None
 
 
 class DeezerClient(Client):
@@ -445,7 +461,7 @@ class DeezerClient(Client):
 
     def __init__(self):
         """Create a DeezerClient."""
-        self.client = deezer.Deezer(accept_language="en-US,en;q=0.5")
+        self.client = deezer.Deezer()
         # self.session = gen_threadsafe_session()
 
         # no login required
@@ -463,7 +479,14 @@ class DeezerClient(Client):
         """
         # TODO: use limit parameter
         try:
-            search_function = getattr(self.client.api, f"search_{media_type}")
+            if media_type == "featured":
+                if query:
+                    search_function = getattr(self.client.api, f"get_editorial_{query}")
+                else:
+                    search_function = self.client.api.get_editorial_releases
+
+            else:
+                search_function = getattr(self.client.api, f"search_{media_type}")
         except AttributeError:
             raise Exception
 
@@ -528,12 +551,12 @@ class DeezerClient(Client):
         dl_info: Dict[str, Any] = {"quality": quality}
 
         track_info = self.client.gw.get_track(meta_id)
-        logger.debug("Track info: %s", pformat(track_info))
+        logger.debug("Track info: %s", track_info)
 
         dl_info["fallback_id"] = safe_get(track_info, "FALLBACK", "SNG_ID")
 
         format_info = get_quality(quality, "deezer")
-        assert isinstance(format_info, tuple)
+        assert isinstance(format_info, tuple)  # for typing
         format_no, format_str = format_info
 
         dl_info["size_to_quality"] = {
@@ -552,7 +575,7 @@ class DeezerClient(Client):
             )
 
         dl_info["url"] = url
-        logger.debug(pformat(dl_info))
+        logger.debug("dl_info %s", dl_info)
         return dl_info
 
     def _get_encrypted_file_url(
@@ -601,7 +624,7 @@ class DeezloaderClient(Client):
     """DeezloaderClient."""
 
     source = "deezer"
-    max_quality = 1
+    max_quality = 2
 
     def __init__(self):
         """Create a DeezloaderClient."""
@@ -658,7 +681,7 @@ class DeezloaderClient(Client):
         return item
 
     @staticmethod
-    def get_file_url(meta_id: Union[str, int], quality: int = 6):
+    def get_file_url(meta_id: Union[str, int], quality: int = 2):
         """Get downloadable url for a track.
 
         :param meta_id: The track ID.
@@ -666,9 +689,9 @@ class DeezloaderClient(Client):
         :param quality:
         :type quality: int
         """
-        quality = min(DEEZER_MAX_Q, quality)
-        url = f"{DEEZER_DL}/{get_quality(quality, 'deezer')}/{DEEZER_BASE}/track/{meta_id}"
-        logger.debug(f"Download url {url}")
+        quality = min(DeezloaderClient.max_quality, quality)
+        url = f"{DEEZER_DL}/{get_quality(quality, 'deezloader')}/{DEEZER_BASE}/track/{meta_id}"
+        logger.debug("Download url %s", url)
         return {"url": url}
 
 
@@ -988,9 +1011,16 @@ class TidalClient(Client):
 
             item["tracks"] = [item["item"] for item in resp["items"]]
         elif media_type == "artist":
-            resp = self._api_request(f"{url}/albums")
-            item["albums"] = resp["items"]
+            logger.debug("filtering eps")
+            album_resp = self._api_request(f"{url}/albums")
+            ep_resp = self._api_request(
+                f"{url}/albums", params={"filter": "EPSANDSINGLES"}
+            )
 
+            item["albums"] = album_resp["items"]
+            item["albums"].extend(ep_resp["items"])
+
+        logger.debug(item)
         return item
 
     def _api_request(self, path: str, params=None) -> dict:
@@ -1007,12 +1037,14 @@ class TidalClient(Client):
         params["countryCode"] = self.country_code
         params["limit"] = 100
         r = self.session.get(f"{TIDAL_BASE}/{path}", params=params)
-        # r.raise_for_status()
+        r.raise_for_status()
         return r.json()
 
     def _get_video_stream_url(self, video_id: str) -> str:
         """Get the HLS video stream url.
 
+        The stream is downloaded using ffmpeg for now.
+
         :param video_id:
         :type video_id: str
         :rtype: str
@@ -1025,16 +1057,18 @@ class TidalClient(Client):
         resp = self._api_request(
             f"videos/{video_id}/playbackinfopostpaywall", params=params
         )
-        stream_url_regex = (
-            r'#EXT-X-STREAM-INF:BANDWIDTH=\d+,AVERAGE-BANDWIDTH=\d+,CODECS="[^"]+"'
-            r",RESOLUTION=\d+x\d+\n(.+)"
-        )
         manifest = json.loads(base64.b64decode(resp["manifest"]).decode("utf-8"))
         available_urls = self.session.get(manifest["urls"][0])
-        url_info = re.findall(stream_url_regex, available_urls.text)
+        available_urls.encoding = "utf-8"
 
-        # highest resolution is last
-        return url_info[-1]
+        STREAM_URL_REGEX = re.compile(
+            r"#EXT-X-STREAM-INF:BANDWIDTH=\d+,AVERAGE-BANDWIDTH=\d+,CODECS=\"(?!jpeg)[^\"]+\",RESOLUTION=\d+x\d+\n(.+)"
+        )
+
+        # Highest resolution is last
+        *_, last_match = STREAM_URL_REGEX.finditer(available_urls.text)
+
+        return last_match.group(1)
 
     def _api_post(self, url, data, auth=None):
         """Post to the Tidal API.
@@ -1052,44 +1086,80 @@ class SoundCloudClient(Client):
 
     source = "soundcloud"
     max_quality = 0
-    logged_in = True
+    logged_in = False
+
+    client_id: str = ""
+    app_version: str = ""
 
     def __init__(self):
         """Create a SoundCloudClient."""
         self.session = gen_threadsafe_session(
             headers={
                 "User-Agent": AGENT,
-                "Host": "api-v2.soundcloud.com",
-                "Origin": "https://soundcloud.com",
-                "Referer": "https://soundcloud.com/",
-                "Sec-Fetch-Dest": "empty",
-                "Sec-Fetch-Mode": "cors",
-                "Sec-Fetch-Site": "same-site",
-                "Sec-GPC": "1",
             }
         )
 
-    def login(self):
-        """Login is not necessary for SoundCloud."""
-        raise NotImplementedError
+    def login(self, **kwargs):
+        self.client_id = kwargs.get("client_id")
+        self.app_version = kwargs.get("app_version")
+        logger.debug("client_id: %s, app_version: %s", self.client_id, self.app_version)
+
+        # if (not self.client_id) or (not self.app_version) or (not self._announce()):
+        if not (self.client_id and self.app_version and self._announce()):
+            logger.debug(
+                "Refreshing client_id=%s and app_version=%s",
+                self.client_id,
+                self.app_version,
+            )
+            self._refresh_tokens()
+
+        self.logged_in = True
+
+    def _announce(self):
+        return self._get("announcements").status_code == 200
+
+    def _refresh_tokens(self):
+        STOCK_URL = "https://soundcloud.com/"
+
+        resp = self.session.get(STOCK_URL)
+        resp.encoding = "utf-8"
+
+        *_, client_id_url_match = re.finditer(
+            r"<script\s+crossorigin\s+src=\"([^\"]+)\"", resp.text
+        )
+        client_id_url = client_id_url_match.group(1)
+
+        self.app_version = re.search(
+            r'<script>window\.__sc_version="(\d+)"</script>', resp.text
+        ).group(1)
+
+        resp2 = self.session.get(client_id_url)
+        self.client_id = re.search(r'client_id:\s*"(\w+)"', resp2.text).group(1)
+
+    def resolve_url(self, url: str) -> dict:
+        resp = self._get(f"resolve?url={url}").json()
+        from pprint import pformat
+
+        logger.debug(pformat(resp))
+        return resp
+
+    def get_tokens(self):
+        return self.client_id, self.app_version
 
     def get(self, id, media_type="track"):
-        """Get metadata for a media type given an id.
+        """Get metadata for a media type given a soundcloud url.
 
         :param id:
         :param media_type:
         """
-        assert media_type in (
+        assert media_type in {
             "track",
             "playlist",
-        ), f"{media_type} not supported"
+        }, f"{media_type} not supported"
 
-        if "http" in str(id):
-            resp, _ = self._get(f"resolve?url={id}")
-        elif media_type == "track":
-            resp, _ = self._get(f"{media_type}s/{id}")
-        else:
-            raise Exception(id)
+        resp = self._get(f"{media_type}s/{id}")
+        resp.raise_for_status()
+        resp = resp.json()
 
         logger.debug(resp)
         return resp
@@ -1112,8 +1182,8 @@ class SoundCloudClient(Client):
             raise Exception
 
         if track["downloadable"] and track["has_downloads_left"]:
-            r = self._get(f"tracks/{track['id']}/download", resp_obj=True)
-            return {"url": r.json()["redirectUri"], "type": "original"}
+            r = self._get(f"tracks/{track['id']}/download")[0]
+            return {"url": r["redirectUri"], "type": "original"}
 
         else:
             url = None
@@ -1125,7 +1195,7 @@ class SoundCloudClient(Client):
 
             assert url is not None
 
-            resp, _ = self._get(url, no_base=True)
+            resp = self._get(url, no_base=True).json()
             return {"url": resp["url"], "type": "mp3"}
 
     def search(self, query: str, media_type="album", limit=50, offset=50):
@@ -1144,10 +1214,19 @@ class SoundCloudClient(Client):
             "offset": offset,
             "linked_partitioning": "1",
         }
-        resp, _ = self._get(f"search/{media_type}s", params=params)
-        return resp
+        result = self._get(f"search/{media_type}s", params=params)
 
-    def _get(self, path, params=None, no_base=False, resp_obj=False):
+        # The response
+        return result.json()
+
+    def _get(
+        self,
+        path,
+        params=None,
+        no_base=False,
+        skip_decode=False,
+        headers=None,
+    ):
         """Send a request to the SoundCloud API.
 
         :param path:
@@ -1159,8 +1238,8 @@ class SoundCloudClient(Client):
         """
         param_arg = params
         params = {
-            "client_id": SOUNDCLOUD_CLIENT_ID,
-            "app_version": SOUNDCLOUD_APP_VERSION,
+            "client_id": self.client_id,
+            "app_version": self.app_version,
             "app_locale": "en",
         }
         if param_arg is not None:
@@ -1171,11 +1250,5 @@ class SoundCloudClient(Client):
         else:
             url = f"{SOUNDCLOUD_BASE}/{path}"
 
-        logger.debug("Fetching url %s", url)
-        logger.debug("Parameters: %s", params)
-        r = self.session.get(url, params=params)
-        logger.debug(r.text)
-        if resp_obj:
-            return r
-
-        return r.json(), r.status_code
+        logger.debug("Fetching url %s with params %s", url, params)
+        return self.session.get(url, params=params, headers=headers)
