@@ -5,9 +5,12 @@ import mutagen.id3 as id3
 AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:83.0) Gecko/20100101 Firefox/83.0"
 
 TIDAL_COVER_URL = "https://resources.tidal.com/images/{uuid}/{width}x{height}.jpg"
-SOUNDCLOUD_CLIENT_ID = "QFciLWLC1GS4P3EZvXIjA3jKhKO5pKB3"
+# Get this from (base64encoded)
+# aHR0cHM6Ly9hLXYyLnNuZGNkbi5jb20vYXNzZXRzLzItYWIxYjg1NjguanM=
+# Don't know if this is a static url yet
+SOUNDCLOUD_CLIENT_ID = "qHsjZaNbdTcABbiIQnVfW07cEPGLNjIh"
 SOUNDCLOUD_USER_ID = "672320-86895-162383-801513"
-SOUNDCLOUD_APP_VERSION = "1626941202"
+SOUNDCLOUD_APP_VERSION = "1630917744"
 
 
 QUALITY_DESC = {
@@ -134,6 +137,7 @@ ALBUM_KEYS = (
     "sampling_rate",
     "container",
     "albumcomposer",
+    "id",
 )
 # TODO: rename these to DEFAULT_FOLDER_FORMAT etc
 FOLDER_FORMAT = (
@@ -152,6 +156,7 @@ TIDAL_Q_MAP = {
 }
 
 DEEZER_MAX_Q = 6
+DEEZER_FEATURED_KEYS = {"releases", "charts", "selection"}
 AVAILABLE_QUALITY_IDS = (0, 1, 2, 3, 4)
 DEEZER_FORMATS = {
     "AAC_64",
