@@ -13,13 +13,23 @@ import re
 import shutil
 import subprocess
 from tempfile import gettempdir
-from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union
+from typing import (
+    Any,
+    Dict,
+    Generator,
+    Iterable,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+)
 
 from click import echo, secho, style
 from mutagen.flac import FLAC, Picture
 from mutagen.id3 import APIC, ID3, ID3NoHeaderError
 from mutagen.mp4 import MP4, MP4Cover
-from pathvalidate import sanitize_filename, sanitize_filepath
+from pathvalidate import sanitize_filepath
 
 from . import converter
 from .clients import Client, DeezloaderClient
@@ -29,6 +39,7 @@ from .constants import (
     FOLDER_FORMAT,
     TRACK_FORMAT,
 )
+from .downloadtools import DownloadPool, DownloadStream
 from .exceptions import (
     InvalidQuality,
     InvalidSourceError,
@@ -39,7 +50,7 @@ from .exceptions import (
 )
 from .metadata import TrackMetadata
 from .utils import (
-    DownloadStream,
+    clean_filename,
     clean_format,
     decrypt_mqa_file,
     downsize_image,
@@ -47,6 +58,7 @@ from .utils import (
     get_container,
     get_cover_urls,
     get_stats_from_quality,
+    get_tqdm_bar,
     safe_get,
     tidal_cover_url,
     tqdm_stream,
@@ -121,6 +133,15 @@ class Media(abc.ABC):
     def downloaded_ids(self, other):
         pass
 
+    @property
+    @abc.abstractmethod
+    def id(self):
+        pass
+
+    @id.setter
+    def id(self, other):
+        pass
+
 
 class Track(Media):
     """Represents a downloadable track.
@@ -194,22 +215,15 @@ class Track(Media):
         self.meta = TrackMetadata(
             track=self.resp, source=source
         )  # meta dict -> TrackMetadata object
-        try:
-            if source == "qobuz":
-                self.cover_url = self.resp["album"]["image"]["large"]
-            elif source == "tidal":
-                self.cover_url = tidal_cover_url(self.resp["album"]["cover"], 320)
-            elif source == "deezer":
-                self.cover_url = self.resp["album"]["cover_medium"]
-            elif source == "soundcloud":
-                self.cover_url = (
-                    self.resp["artwork_url"] or self.resp["user"].get("avatar_url")
-                ).replace("large", "t500x500")
-            else:
-                raise InvalidSourceError(source)
-        except KeyError:
-            logger.debug("No cover found")
-            self.cover_url = None
+
+        # Because the cover urls are not parsed when only the track metadata
+        # is loaded, we need to do this ourselves
+
+        # Default to large if chosen size is not available
+        self.cover_url = self.meta.cover_urls.get(
+            kwargs.get("embed_cover_size", "large"),
+            self.meta.cover_urls.get("large"),
+        )
 
     def _prepare_download(self, **kwargs):
         """Do preprocessing before downloading items.
@@ -232,13 +246,16 @@ class Track(Media):
                 clean_format(
                     kwargs.get("folder_format", FOLDER_FORMAT),
                     self.meta.get_album_formatter(self.quality),
+                    restrict=kwargs.get("restrict_filenames", False),
                 ),
             )
 
         self.file_format = kwargs.get("track_format", TRACK_FORMAT)
 
         self.folder = sanitize_filepath(self.folder, platform="auto")
-        self.format_final_path()  # raises: ItemExists
+        self.format_final_path(
+            restrict=kwargs.get("restrict_filenames", False)
+        )  # raises: ItemExists
 
         os.makedirs(self.folder, exist_ok=True)
 
@@ -253,7 +270,7 @@ class Track(Media):
 
         self.path = os.path.join(gettempdir(), f"{hash(self.id)}_{self.quality}.tmp")
 
-    def download(
+    def download(  # noqa
         self,
         quality: int = 3,
         parent_folder: str = "StreamripDownloads",
@@ -269,7 +286,7 @@ class Track(Media):
         :param progress_bar: turn on/off progress bar
         :type progress_bar: bool
         """
-        if not self.part_of_tracklist:
+        if not self.part_of_tracklist and not self.client.source == "soundcloud":
             secho(f"Downloading {self!s}\n", bold=True)
 
         self._prepare_download(
@@ -288,7 +305,7 @@ class Track(Media):
         try:
             dl_info = self.client.get_file_url(url_id, self.quality)
         except Exception as e:
-            # raise NonStreamable(repr(e))
+            logger.debug(repr(e))
             raise NonStreamable(e)
 
         if self.client.source == "qobuz":
@@ -340,7 +357,9 @@ class Track(Media):
             if self.quality != stream_quality:
                 # The chosen quality is not available
                 self.quality = stream_quality
-                self.format_final_path()  # If the extension is different
+                self.format_final_path(
+                    restrict=kwargs.get("restrict_filenames", False)
+                )  # If the extension is different
 
             with open(self.path, "wb") as file:
                 for chunk in tqdm_stream(stream, desc=self._progress_desc):
@@ -421,21 +440,33 @@ class Track(Media):
         """
         logger.debug("dl_info: %s", dl_info)
         if dl_info["type"] == "mp3":
+            import m3u8
+            import requests
+
+            parsed_m3u = m3u8.loads(requests.get(dl_info["url"]).text)
             self.path += ".mp3"
-            # convert hls stream to mp3
-            subprocess.call(
-                [
-                    "ffmpeg",
-                    "-i",
-                    dl_info["url"],
-                    "-c",
-                    "copy",
-                    "-y",
-                    self.path,
-                    "-loglevel",
-                    "fatal",
-                ]
-            )
+
+            with DownloadPool(segment.uri for segment in parsed_m3u.segments) as pool:
+
+                bar = get_tqdm_bar(len(pool), desc=self._progress_desc, unit="Chunk")
+
+                def update_tqdm_bar():
+                    bar.update(1)
+
+                pool.download(callback=update_tqdm_bar)
+                subprocess.call(
+                    (
+                        "ffmpeg",
+                        "-i",
+                        f"concat:{'|'.join(pool.files)}",
+                        "-acodec",
+                        "copy",
+                        "-loglevel",
+                        "panic",
+                        self.path,
+                    )
+                )
+
         elif dl_info["type"] == "original":
             _quick_download(dl_info["url"], self.path, desc=self._progress_desc)
 
@@ -466,7 +497,7 @@ class Track(Media):
     def download_cover(self, width=999999, height=999999):
         """Download the cover art, if cover_url is given."""
         self.cover_path = os.path.join(gettempdir(), f"cover{hash(self.cover_url)}.jpg")
-        logger.debug(f"Downloading cover from {self.cover_url}")
+        logger.debug("Downloading cover from %s", self.cover_url)
 
         if not os.path.exists(self.cover_path):
             _cover_download(self.cover_url, self.cover_path)
@@ -475,7 +506,7 @@ class Track(Media):
             logger.debug("Cover already exists, skipping download")
             raise ItemExists(self.cover_path)
 
-    def format_final_path(self) -> str:
+    def format_final_path(self, restrict: bool = False) -> str:
         """Return the final filepath of the downloaded file.
 
         This uses the `get_formatter` method of TrackMetadata, which returns
@@ -484,7 +515,7 @@ class Track(Media):
         """
         formatter = self.meta.get_formatter(max_quality=self.quality)
         logger.debug("Track meta formatter %s", formatter)
-        filename = clean_format(self.file_format, formatter)
+        filename = clean_format(self.file_format, formatter, restrict=restrict)
         self.final_path = os.path.join(self.folder, filename)[:250].strip() + ext(
             self.quality, self.client.source
         )
@@ -549,11 +580,12 @@ class Track(Media):
             cover_url=cover_url,
         )
 
-    def tag(
+    def tag(  # noqa
         self,
         album_meta: dict = None,
         cover: Union[Picture, APIC, MP4Cover] = None,
         embed_cover: bool = True,
+        exclude_tags: Optional[Sequence] = None,
         **kwargs,
     ):
         """Tag the track using the stored metadata.
@@ -620,7 +652,10 @@ class Track(Media):
                 raise InvalidQuality(f'Invalid quality: "{self.quality}"')
 
         # automatically generate key, value pairs based on container
-        tags = self.meta.tags(self.container)
+        tags = self.meta.tags(
+            self.container,
+            set(exclude_tags) if exclude_tags is not None else None,
+        )
         for k, v in tags:
             logger.debug("Setting %s tag to %s", k, v)
             audio[k] = v
@@ -687,7 +722,7 @@ class Track(Media):
             exit()
 
         if not hasattr(self, "final_path"):
-            self.format_final_path()
+            self.format_final_path(kwargs.get("restrict_filenames", False))
 
         if not os.path.isfile(self.path):
             logger.info("File %s does not exist. Skipping conversion.", self.path)
@@ -720,14 +755,15 @@ class Track(Media):
 
         :rtype: str
         """
-        if hasattr(self, "meta"):
+        try:
             _title = self.meta.title
-            if self.meta.explicit:
-                _title = f"{_title} (Explicit)"
-            return _title
-        else:
+        except AttributeError:
             raise Exception("Track must be loaded before accessing title")
 
+        if self.meta.explicit:
+            _title = f"{_title} (Explicit)"
+        return _title
+
     def get(self, *keys, default=None) -> Any:
         """Safe get method that allows for layered access.
 
@@ -783,6 +819,9 @@ class Track(Media):
 class Video(Media):
     """Only for Tidal."""
 
+    id = None
+    downloaded_ids: set = set()
+
     def __init__(self, client: Client, id: str, **kwargs):
         """Initialize a Video object.
 
@@ -794,41 +833,71 @@ class Video(Media):
         """
         self.id = id
         self.client = client
-        self.title = kwargs.get("title", "MusicVideo")
-        self.explicit = kwargs.get("explicit", False)
-        self.tracknumber = kwargs.get("tracknumber", None)
 
     def load_meta(self, **kwargs):
         """Given an id at contruction, get the metadata of the video."""
         resp = self.client.get(self.id, "video")
         self.title = resp["title"]
         self.explicit = resp["explicit"]
+        self.tracknumber = resp["trackNumber"]
 
     def download(self, **kwargs):
         """Download the Video.
 
         :param kwargs:
         """
-        secho(
-            f"Downloading {self.title} (Video). This may take a while.",
-            fg="blue",
-        )
+
+        if not kwargs.get("download_videos", True):
+            return
+
+        import m3u8
+        import requests
+
+        # secho(
+        #     f"Downloading {self.title} (Video). This may take a while.",
+        #     fg="blue",
+        # )
 
         self.parent_folder = kwargs.get("parent_folder", "StreamripDownloads")
         url = self.client.get_file_url(self.id, video=True)
-        # it's more convenient to have ffmpeg download the hls
-        command = [
-            "ffmpeg",
-            "-i",
-            url,
-            "-c",
-            "copy",
-            "-loglevel",
-            "panic",
-            self.path,
-        ]
-        p = subprocess.Popen(command)
-        p.wait()  # remove this?
+
+        parsed_m3u = m3u8.loads(requests.get(url).text)
+        # Asynchronously download the streams
+
+        with DownloadPool(segment.uri for segment in parsed_m3u.segments) as pool:
+            bar = get_tqdm_bar(len(pool), desc=self._progress_desc, unit="Chunk")
+
+            def update_tqdm_bar():
+                bar.update(1)
+
+            pool.download(callback=update_tqdm_bar)
+
+            # Put the filenames in a tempfile that ffmpeg
+            # can read from
+            file_list_path = os.path.join(gettempdir(), "__streamrip_video_files")
+            with open(file_list_path, "w") as file_list:
+                text = "\n".join(f"file '{path}'" for path in pool.files)
+                file_list.write(text)
+
+            # Use ffmpeg to concat the files
+            subprocess.call(
+                (
+                    "ffmpeg",
+                    "-f",
+                    "concat",
+                    "-safe",
+                    "0",
+                    "-i",
+                    file_list_path,
+                    "-c",
+                    "copy",
+                    "-loglevel",
+                    "panic",
+                    self.path,
+                )
+            )
+
+            os.remove(file_list_path)
 
     def tag(self, *args, **kwargs):
         """Return False.
@@ -867,6 +936,10 @@ class Video(Media):
         """
         pass
 
+    @property
+    def _progress_desc(self) -> str:
+        return style(f"Video {self.tracknumber:02}", fg="blue")
+
     @property
     def path(self) -> str:
         """Get path to download the mp4 file.
@@ -912,6 +985,8 @@ class Video(Media):
 class YoutubeVideo(Media):
     """Dummy class implemented for consistency with the Media API."""
 
+    id = None
+
     class DummyClient:
         """Used because YouTube downloads use youtube-dl, not a client."""
 
@@ -948,6 +1023,7 @@ class YoutubeVideo(Media):
         filename_formatter = "%(track_number)s.%(track)s.%(container)s"
         filename = os.path.join(parent_folder, filename_formatter)
 
+        assert isinstance(self.id, str)
         p = subprocess.Popen(
             [
                 "youtube-dl",
@@ -1000,6 +1076,12 @@ class YoutubeVideo(Media):
         """
         pass
 
+    def convert(self, *args, **kwargs):
+        raise NotImplementedError
+
+    def __repr__(self, *args, **kwargs):
+        return f"YoutubeVideo({self.url})"
+
     def __bool__(self):
         """Return True."""
         return True
@@ -1008,6 +1090,8 @@ class YoutubeVideo(Media):
 class Booklet:
     """Only for Qobuz."""
 
+    id = None
+
     def __init__(self, resp: dict):
         """Initialize from the `goodies` field of the Qobuz API response.
 
@@ -1031,9 +1115,9 @@ class Booklet:
         :type parent_folder: str
         :param kwargs:
         """
-        filepath = os.path.join(
-            parent_folder, f"{sanitize_filename(self.description)}.pdf"
-        )
+        fn = clean_filename(self.description, restrict=kwargs.get("restrict_filenames"))
+        filepath = os.path.join(parent_folder, f"{fn}.pdf")
+
         _quick_download(self.url, filepath, "Booklet")
 
     def type(self) -> str:
@@ -1057,6 +1141,7 @@ class Tracklist(list):
     the tracklist.
     """
 
+    id = None
     # anything not in parentheses or brackets
     essence_regex = re.compile(r"([^\(]+)(?:\s*[\(\[][^\)][\)\]])*")
 
@@ -1191,7 +1276,7 @@ class Tracklist(list):
                     kwargs["sampling_rate"],
                 )
             else:
-                logger.debug(f"Downsampling to {sr/1000}kHz")
+                logger.debug("Downsampling to %skHz", sr / 1000)
 
         for track in self:
             track.convert(codec, **kwargs)
@@ -1331,16 +1416,15 @@ class Tracklist(list):
 class Album(Tracklist, Media):
     """Represents a downloadable album.
 
-    Usage:
-
     >>> resp = client.get('fleetwood mac rumours', 'album')
     >>> album = Album.from_api(resp['items'][0], client)
     >>> album.load_meta()
     >>> album.download()
+
     """
 
     downloaded_ids: set = set()
-    id: str
+    id = None
 
     def __init__(self, client: Client, **kwargs):
         """Create a new Album object.
@@ -1382,7 +1466,7 @@ class Album(Tracklist, Media):
         if not self.get("streamable", False):
             raise NonStreamable(f"This album is not streamable ({self.id} ID)")
 
-        self._load_tracks(resp)
+        self._load_tracks(resp, kwargs.get("download_videos", True))
         self.loaded = True
 
     @classmethod
@@ -1409,52 +1493,37 @@ class Album(Tracklist, Media):
         self.folder_format = kwargs.get("folder_format", FOLDER_FORMAT)
         self.quality = min(kwargs.get("quality", 3), self.client.max_quality)
 
-        self.folder = self._get_formatted_folder(
-            kwargs.get("parent_folder", "StreamripDownloads"), self.quality
-        )
+        self.container = get_container(self.quality, self.client.source)
+        # necessary to format the folder
+        if self.container in ("AAC", "MP3"):
+            # lossy codecs don't have these metrics
+            self.bit_depth = self.sampling_rate = None
+
+        parent_folder = kwargs.get("parent_folder", "StreamripDownloads")
+        if self.folder_format:
+            self.folder = self._get_formatted_folder(
+                parent_folder, restrict=kwargs.get("restrict_filenames", False)
+            )
+        else:
+            self.folder = parent_folder
+
         os.makedirs(self.folder, exist_ok=True)
 
         self.download_message()
 
-        # choose optimal cover size and download it
-        secho("Downloading cover art", bold=True)
-        cover_path = os.path.join(gettempdir(), f"cover_{hash(self)}.jpg")
-        embed_cover_size = kwargs.get("embed_cover_size", "large")
-
-        assert (
-            embed_cover_size in self.cover_urls
-        ), f"Invalid cover size. Must be in {self.cover_urls.keys()}"
-
-        embed_cover_url = self.cover_urls[embed_cover_size]
-        if not os.path.exists(cover_path):
-            cover_url = (
-                embed_cover_url
-                if embed_cover_url is None
-                else tuple(filter(None, self.cover_urls.values()))[0]
-            )
-
-            _cover_download(cover_url, cover_path)
-
-        hires_cov_path = os.path.join(self.folder, "cover.jpg")
-        if kwargs.get("keep_hires_cover", True) and not os.path.exists(hires_cov_path):
-            _cover_download(self.cover_urls["original"], hires_cov_path)
-
-        cover_size = os.path.getsize(cover_path)
-        if cover_size > FLAC_MAX_BLOCKSIZE:  # 16.77 MB
-            secho(
-                "Downgrading embedded cover size, too large ({cover_size}).",
-                fg="bright_yellow",
-            )
-            # large is about 600x600px which is guaranteed < 16.7 MB
-            _cover_download(self.cover_urls["large"], cover_path)
-
-        downsize_image(
-            cover_path,
-            kwargs.get("max_artwork_width", 999999),
-            kwargs.get("max_artwork_height", 999999),
+        cover_path = _choose_and_download_cover(
+            self.cover_urls,
+            kwargs.get("embed_cover_size", "large"),
+            self.folder,
+            kwargs.get("keep_hires_cover", True),
+            (
+                kwargs.get("max_artwork_width", 999999),
+                kwargs.get("max_artwork_height", 999999),
+            ),
         )
 
         if kwargs.get("embed_cover", True):  # embed by default
+            logger.debug("Getting cover_obj from %s", cover_path)
             # container generated when formatting folder name
             self.cover_obj = self.get_cover_obj(
                 cover_path, self.container, self.client.source
@@ -1483,7 +1552,11 @@ class Album(Tracklist, Media):
         :rtype: bool
         """
         logger.debug("Downloading track to %s", self.folder)
-        if self.disctotal > 1 and isinstance(item, Track):
+        if (
+            self.disctotal > 1
+            and isinstance(item, Track)
+            and kwargs.get("folder_format")
+        ):
             disc_folder = os.path.join(self.folder, f"Disc {item.meta.discnumber}")
             kwargs["parent_folder"] = disc_folder
         else:
@@ -1499,6 +1572,7 @@ class Album(Tracklist, Media):
             item.tag(
                 cover=self.cover_obj,
                 embed_cover=kwargs.get("embed_cover", True),
+                exclude_tags=kwargs.get("exclude_tags"),
             )
 
         self.downloaded_ids.add(item.id)
@@ -1515,7 +1589,7 @@ class Album(Tracklist, Media):
         meta.id = resp["id"]
         return meta
 
-    def _load_tracks(self, resp):
+    def _load_tracks(self, resp, download_videos: bool = True):
         """Load the tracks into self from an API response.
 
         This uses a classmethod to convert an item into a Track object, which
@@ -1524,7 +1598,8 @@ class Album(Tracklist, Media):
         logging.debug("Loading %d tracks to album", self.tracktotal)
         for track in _get_tracklist(resp, self.client.source):
             if track.get("type") == "Music Video":
-                self.append(Video.from_album_meta(track, self.client))
+                if download_videos:
+                    self.append(Video.from_album_meta(track, self.client))
             else:
                 self.append(
                     Track.from_album_meta(
@@ -1548,6 +1623,8 @@ class Album(Tracklist, Media):
             )
             if stat1 is not None and stat2 is not None
         )
+        logger.debug("Sampling rate, bit depth = %s", stats)
+
         if not stats:
             stats = (None, None)
 
@@ -1566,7 +1643,7 @@ class Album(Tracklist, Media):
         logger.debug("Formatter: %s", fmt)
         return fmt
 
-    def _get_formatted_folder(self, parent_folder: str, quality: int) -> str:
+    def _get_formatted_folder(self, parent_folder: str, restrict: bool = False) -> str:
         """Generate the folder name for this album.
 
         :param parent_folder:
@@ -1575,13 +1652,12 @@ class Album(Tracklist, Media):
         :type quality: int
         :rtype: str
         """
-        # necessary to format the folder
-        self.container = get_container(quality, self.client.source)
-        if self.container in ("AAC", "MP3"):
-            # lossy codecs don't have these metrics
-            self.bit_depth = self.sampling_rate = None
 
-        formatted_folder = clean_format(self.folder_format, self._get_formatter())
+        formatted_folder = clean_format(
+            self.folder_format,
+            self._get_formatter(),
+            restrict=restrict,
+        )
 
         return os.path.join(parent_folder, formatted_folder)
 
@@ -1708,7 +1784,7 @@ class Playlist(Tracklist, Media):
 
         elif self.client.source == "tidal":
             self.name = self.meta["title"]
-            self.image = tidal_cover_url(self.meta["image"], 640)
+            self.image = tidal_cover_url(self.meta["image"] or self.meta['squareImage'], 640)
             self.creator = safe_get(self.meta, "creator", "name", default="TIDAL")
 
             tracklist = self.meta["tracks"]
@@ -1761,11 +1837,14 @@ class Playlist(Tracklist, Media):
                     )
                 )
 
-        logger.debug(f"Loaded {len(self)} tracks from playlist {self.name}")
+        logger.debug("Loaded %d tracks from playlist %s", len(self), self.name)
 
     def _prepare_download(self, parent_folder: str = "StreamripDownloads", **kwargs):
-        fname = sanitize_filename(self.name)
-        self.folder = os.path.join(parent_folder, fname)
+        if kwargs.get("folder_format"):
+            fname = clean_filename(self.name, kwargs.get("restrict_filenames", False))
+            self.folder = os.path.join(parent_folder, fname)
+        else:
+            self.folder = parent_folder
 
         # Used for safe concurrency with tracknumbers instead of an object
         # level that stores an index
@@ -1778,7 +1857,6 @@ class Playlist(Tracklist, Media):
         kwargs["parent_folder"] = self.folder
         if self.client.source == "soundcloud":
             item.load_meta()
-            secho(f"Downloading {item!s}", fg="blue")
 
         if kwargs.get("set_playlist_to_album", False):
             item.meta.album = self.name
@@ -1790,18 +1868,10 @@ class Playlist(Tracklist, Media):
 
         item.download(**kwargs)
 
-        item.tag(embed_cover=kwargs.get("embed_cover", True))
-
-        # if playlist_to_album and self.client.source == "deezer":
-        #     # Because Deezer tracks come pre-tagged, the `set_playlist_to_album`
-        #     # option is never set. Here, we manually do this
-        #     from mutagen.flac import FLAC
-
-        #     audio = FLAC(item.path)
-        #     audio["ALBUM"] = self.name
-        #     audio["ALBUMARTIST"] = self.creator
-        #     audio["TRACKNUMBER"] = f"{item['tracknumber']:02}"
-        #     audio.save()
+        item.tag(
+            embed_cover=kwargs.get("embed_cover", True),
+            exclude_tags=kwargs.get("exclude_tags"),
+        )
 
         self.downloaded_ids.add(item.id)
 
@@ -1877,6 +1947,7 @@ class Artist(Tracklist, Media):
     >>> artist.download()
     """
 
+    id = None
     downloaded_ids: set = set()
 
     def __init__(self, client: Client, **kwargs):
@@ -1888,6 +1959,7 @@ class Artist(Tracklist, Media):
         :param kwargs:
         """
         self.client = client
+        self.downloaded_ids = set()
 
         for k, v in kwargs.items():
             setattr(self, k, v)
@@ -1953,12 +2025,15 @@ class Artist(Tracklist, Media):
         :param kwargs:
         :rtype: Iterable
         """
-        folder = sanitize_filename(self.name)
-        self.folder = os.path.join(parent_folder, folder)
+        if kwargs.get("folder_format"):
+            folder = clean_filename(self.name, kwargs.get("restrict_filenames", False))
+            self.folder = os.path.join(parent_folder, folder)
+        else:
+            self.folder = parent_folder
 
         logger.debug("Artist folder: %s", folder)
-        logger.debug(f"Length of tracklist {len(self)}")
-        logger.debug(f"Filters: {filters}")
+        logger.debug("Length of tracklist %d", len(self))
+        logger.debug("Filters: %s", filters)
 
         final: Iterable
         if "repeats" in filters:
@@ -1994,10 +2069,14 @@ class Artist(Tracklist, Media):
 
         kwargs.pop("parent_folder")
         # always an Album
-        item.download(
-            parent_folder=self.folder,
-            **kwargs,
-        )
+        try:
+            item.download(
+                parent_folder=self.folder,
+                **kwargs,
+            )
+        except PartialFailure:
+            pass
+
         self.downloaded_ids.update(item.downloaded_ids)
 
     @property
@@ -2161,6 +2240,8 @@ class Artist(Tracklist, Media):
 class Label(Artist):
     """Represents a downloadable Label."""
 
+    id = None
+
     def load_meta(self, **kwargs):
         """Load metadata given an id."""
         assert self.client.source == "qobuz", "Label source must be qobuz"
@@ -2212,3 +2293,51 @@ def _quick_download(url: str, path: str, desc: str = None):
 
 def _cover_download(url: str, path: str):
     _quick_download(url, path, style("Cover", fg="blue"))
+
+
+def _choose_and_download_cover(
+    cover_urls: dict,
+    preferred_size: str,
+    directory: str,
+    keep_hires_cover: bool = True,
+    downsize: Tuple[int, int] = (999999, 999999),
+) -> str:
+    # choose optimal cover size and download it
+    temp_cover_path = os.path.join(
+        gettempdir(), f"cover_{hash(cover_urls.values())}.jpg"
+    )
+    secho(f"Downloading cover art ({preferred_size})", bold=True)
+
+    assert (
+        preferred_size in cover_urls
+    ), f"Invalid cover size. Must be in {cover_urls.keys()}"
+
+    embed_cover_url = cover_urls[preferred_size]
+    logger.debug("Chosen cover url: %s", embed_cover_url)
+    if not os.path.exists(temp_cover_path):
+        # Sometimes a size isn't available. When this is the case, find
+        # the first `not None` url.
+        if embed_cover_url is None:
+            embed_cover_url = next(filter(None, cover_urls.values()))
+
+        logger.debug("Downloading cover from url %s", embed_cover_url)
+
+        _cover_download(embed_cover_url, temp_cover_path)
+
+    hires_cov_path = os.path.join(directory, "cover.jpg")
+    if keep_hires_cover and not os.path.exists(hires_cov_path):
+        logger.debug("Downloading hires cover")
+        _cover_download(cover_urls["original"], hires_cov_path)
+
+    cover_size = os.path.getsize(temp_cover_path)
+    if cover_size > FLAC_MAX_BLOCKSIZE:  # 16.77 MB
+        secho(
+            "Downgrading embedded cover size, too large ({cover_size}).",
+            fg="bright_yellow",
+        )
+        # large is about 600x600px which is guaranteed < 16.7 MB
+        _cover_download(cover_urls["large"], temp_cover_path)
+
+    downsize_image(temp_cover_path, *downsize)
+
+    return temp_cover_path
