@@ -49,6 +49,39 @@ class TrackMetadata:
         * disctotal
     """
 
+    albumartist: str
+    composer: Optional[str] = None
+    albumcomposer: Optional[str] = None
+    comment: Optional[str] = None
+    description: Optional[str] = None
+    purchase_date: Optional[str] = None
+    date: Optional[str] = None
+    grouping: Optional[str] = None
+    lyrics: Optional[str] = None
+    encoder: Optional[str] = None
+    compilation: Optional[str] = None
+    cover: Optional[str] = None
+    tracktotal: Optional[int] = None
+    tracknumber: Optional[int] = None
+    discnumber: Optional[int] = None
+    disctotal: Optional[int] = None
+
+    # not included in tags
+    explicit: Optional[bool] = False
+    quality: Optional[int] = None
+    sampling_rate: Optional[int] = None
+    bit_depth: Optional[int] = None
+    booklets = None
+    cover_urls = Optional[OrderedDict]
+    work: Optional[str]
+    id: Optional[str]
+
+    # Internals
+    _artist: Optional[str] = None
+    _copyright: Optional[str] = None
+    _genres: Optional[Iterable] = None
+    _title: Optional[str]
+
     def __init__(
         self,
         track: Optional[Union[TrackMetadata, dict]] = None,
@@ -63,40 +96,7 @@ class TrackMetadata:
         :type album: Optional[dict]
         """
         # embedded information
-        self.title: str
-        self.album: str
-        self.albumartist: str
-        self.composer: Optional[str] = None
-        self.albumcomposer: Optional[str] = None
-        self.comment: Optional[str] = None
-        self.description: Optional[str] = None
-        self.purchase_date: Optional[str] = None
-        self.date: Optional[str] = None
-        self.grouping: Optional[str] = None
-        self.lyrics: Optional[str] = None
-        self.encoder: Optional[str] = None
-        self.compilation: Optional[str] = None
-        self.cover: Optional[str] = None
-        self.tracktotal: Optional[int] = None
-        self.tracknumber: Optional[int] = None
-        self.discnumber: Optional[int] = None
-        self.disctotal: Optional[int] = None
-
-        # not included in tags
-        self.explicit: Optional[bool] = False
-        self.quality: Optional[int] = None
-        self.sampling_rate: Optional[int] = None
-        self.bit_depth: Optional[int] = None
-        self.booklets = None
-        self.cover_urls = Optional[OrderedDict]
-        self.work: Optional[str]
-        self.id: Optional[str]
-
-        # Internals
-        self._artist: Optional[str] = None
-        self._copyright: Optional[str] = None
-        self._genres: Optional[Iterable] = None
-
+        # TODO: add this to static attrs
         self.__source = source
 
         if isinstance(track, TrackMetadata):
@@ -130,10 +130,15 @@ class TrackMetadata:
             # Tags
             self.album = resp.get("title", "Unknown Album")
             self.tracktotal = resp.get("tracks_count", 1)
-            self.genre = resp.get("genres_list") or resp.get("genre")
+            self.genre = resp.get("genres_list") or resp.get("genre") or []
             self.date = resp.get("release_date_original") or resp.get("release_date")
             self.copyright = resp.get("copyright")
-            self.albumartist = safe_get(resp, "artist", "name")
+
+            if artists := resp.get("artists"):
+                self.albumartist = ", ".join(a["name"] for a in artists)
+            else:
+                self.albumartist = safe_get(resp, "artist", "name")
+
             self.albumcomposer = safe_get(resp, "composer", "name")
             self.label = resp.get("label")
             self.description = resp.get("description")
@@ -157,6 +162,7 @@ class TrackMetadata:
             self.sampling_rate = resp.get("maximum_sampling_rate")
             self.quality = get_quality_id(self.bit_depth, self.sampling_rate)
             self.booklets = resp.get("goodies")
+            self.id = resp.get("id")
 
             if self.sampling_rate is not None:
                 self.sampling_rate *= 1000
@@ -178,6 +184,7 @@ class TrackMetadata:
             # 80, 160, 320, 640, 1280
             self.cover_urls = get_cover_urls(resp, self.__source)
             self.streamable = resp.get("allowStreaming", False)
+            self.id = resp.get("id")
 
             if q := resp.get("audioQuality"):  # for album entries in single tracks
                 self._get_tidal_quality(q)
@@ -193,14 +200,16 @@ class TrackMetadata:
             self.albumartist = safe_get(resp, "artist", "name")
             self.label = resp.get("label")
             self.url = resp.get("link")
+            self.explicit = bool(resp.get("parental_warning"))
 
             # not embedded
-            self.explicit = bool(resp.get("parental_warning"))
             self.quality = 2
-            self.bit_depth = None
+            self.bit_depth = 16
+            self.sampling_rate = 44100
+
             self.cover_urls = get_cover_urls(resp, self.__source)
-            self.sampling_rate = None
             self.streamable = True
+            self.id = resp.get("id")
 
         elif self.__source == "soundcloud":
             raise NotImplementedError
@@ -248,6 +257,7 @@ class TrackMetadata:
             self.tracknumber = 0
             self.tracktotal = 0
             self.quality = 0
+            self.cover_urls = get_cover_urls(track, "soundcloud")
 
         else:
             raise ValueError(self.__source)
@@ -275,6 +285,20 @@ class TrackMetadata:
             self.bit_depth = 24 if self.get("quality") == 3 else 16
             self.sampling_rate = 44100
 
+    @property
+    def title(self) -> Optional[str]:
+        if not hasattr(self, "_title"):
+            return None
+
+        if self.explicit:
+            return f"{self._title} (Explicit)"
+
+        return self._title
+
+    @title.setter
+    def title(self, new_title):
+        self._title = new_title
+
     @property
     def album(self) -> str:
         """Return the album of the track.
@@ -345,7 +369,9 @@ class TrackMetadata:
                 genres: Iterable = re.findall(r"([^\u2192\/]+)", "/".join(self._genres))
                 genres = set(genres)
             elif self.__source == "deezer":
-                genres = ", ".join(g["name"] for g in self._genres)
+                genres = (g["name"] for g in self._genres)
+            else:
+                raise Exception
 
             return ", ".join(genres)
 
@@ -382,7 +408,7 @@ class TrackMetadata:
         return None
 
     @copyright.setter
-    def copyright(self, val: str):
+    def copyright(self, val: Optional[str]):
         """Set the internal copyright variable to the given value.
 
         Only formatted when requested.
@@ -434,7 +460,7 @@ class TrackMetadata:
         formatter["sampling_rate"] /= 1000
         return formatter
 
-    def tags(self, container: str = "flac") -> Generator:
+    def tags(self, container: str = "flac", exclude: Optional[set] = None) -> Generator:
         """Create a generator of key, value pairs for use with mutagen.
 
         The *_KEY dicts are organized in the format:
@@ -456,41 +482,52 @@ class TrackMetadata:
         :type container: str
         :rtype: Generator
         """
+        if exclude is None:
+            exclude = set()
+        logger.debug("Excluded tags: %s", exclude)
+
         container = container.lower()
         if container in ("flac", "vorbis"):
-            return self.__gen_flac_tags()
+            return self.__gen_flac_tags(exclude)
         if container in ("mp3", "id3"):
-            return self.__gen_mp3_tags()
+            return self.__gen_mp3_tags(exclude)
         if container in ("alac", "m4a", "mp4", "aac"):
-            return self.__gen_mp4_tags()
+            return self.__gen_mp4_tags(exclude)
 
         raise InvalidContainerError(f"Invalid container {container}")
 
-    def __gen_flac_tags(self) -> Generator:
+    def __gen_flac_tags(self, exclude: set) -> Generator:
         """Generate key, value pairs to tag FLAC files.
 
         :rtype: Tuple[str, str]
         """
         for k, v in FLAC_KEY.items():
+            logger.debug("attr: %s", k)
+            if k in exclude:
+                continue
+
             tag = getattr(self, k)
             if tag:
-                if k in (
+                if k in {
                     "tracknumber",
                     "discnumber",
                     "tracktotal",
                     "disctotal",
-                ):
+                }:
                     tag = f"{int(tag):02}"
 
                 logger.debug("Adding tag %s: %s", v, tag)
                 yield (v, str(tag))
 
-    def __gen_mp3_tags(self) -> Generator:
+    def __gen_mp3_tags(self, exclude: set) -> Generator:
         """Generate key, value pairs to tag MP3 files.
 
         :rtype: Tuple[str, str]
         """
         for k, v in MP3_KEY.items():
+            if k in exclude:
+                continue
+
             if k == "tracknumber":
                 text = f"{self.tracknumber}/{self.tracktotal}"
             elif k == "discnumber":
@@ -501,12 +538,15 @@ class TrackMetadata:
             if text is not None and v is not None:
                 yield (v.__name__, v(encoding=3, text=text))
 
-    def __gen_mp4_tags(self) -> Generator:
+    def __gen_mp4_tags(self, exclude: set) -> Generator:
         """Generate key, value pairs to tag ALAC or AAC files.
 
         :rtype: Tuple[str, str]
         """
         for k, v in MP4_KEY.items():
+            if k in exclude:
+                continue
+
             if k == "tracknumber":
                 text = [(self.tracknumber, self.tracktotal)]
             elif k == "discnumber":
